package eu.clarin.cmdi.oai.harvester.impl;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.stream.XMLStreamException;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.CoreProtocolPNames;
import org.apache.http.params.HttpParams;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.MutableDateTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;

import eu.clarin.cmdi.oai.harvester.HarvestJob.State;
import eu.clarin.cmdi.oai.harvester.HarvesterException;
import eu.clarin.cmdi.oai.harvester.MetadataFormat;
import eu.clarin.cmdi.oai.harvester.Repository;
import eu.clarin.cmdi.oai.harvester.Repository.DeletedNotion;
import eu.clarin.cmdi.oai.harvester.Repository.Granularity;
import eu.clarin.cmdi.oai.harvester.impl.HarvestJobImpl.Task;

public class HarvestWorker {
    private static final String OAI_NS = "http://www.openarchives.org/OAI/2.0/";
    private static final String ARG_METADATA_PREFIX = "metadataPrefix";
    private static final String ARG_RESUMPTION_TOKEN = "resumptionToken";
    private static final String PROTOCOL_VERSION = "2.0";
    private static final String DELETED_NOTION_NO = "no";
    private static final String DELETED_NOTION_PERSISTENT = "persistent";
    private static final String DELETED_NOTION_TRANSIENT = "transient";
    private static final String GRANULARITY_DAYS = "YYYY-MM-DD";
    private static final String GRANULARITY_SECONDS = "YYYY-MM-DDThh:mm:ssZ";
    private static final String COMPRESSION_DEFLATE = "deflate";
    private static final String COMPRESSION_GZIP = "gzip";
    private static final String VERB_IDENTIFY = "Identify";
    private static final String VERB_LIST_METADATAFORMATS =
        "ListMetadataFormats";
    private static final String VERB_LIST_RECORDS = "ListRecords";
    private static final String HEADER_ACCEPT_ENCODING = "Accept-Encoding";
    private static final String HEADER_CONNECTION = "Connection";
    private static final String[] DATEFORMATS_DAYS =
        { "yyyy'-'MM'-'dd" };
    private static final String[] DATEFORMATS_FULL =
        { "yyyy'-'MM'-'dd'T'HH':'mm':'ss'Z'", "yyyy'-'MM'-'dd" };
    private final AbstractHarvester harvester;
    private final DefaultHttpClient client;
    private final RequestUriBuilder uriBuilder;
    private final HeaderImpl header;
    private final XmlStreamReaderProxy readerWrapper;
    private long delay;

    HarvestWorker(AbstractHarvester harvester) {
        this.harvester = harvester;
        // http client
        client = new DefaultHttpClient();
        client.getParams().setParameter(CoreProtocolPNames.USER_AGENT,
                    "eu.clarin.cmdi.oai.Harvester/0.0.1");
        // request stuff
        uriBuilder = new RequestUriBuilder();
        // other stuff
        this.header = new HeaderImpl();
        this.readerWrapper = new XmlStreamReaderProxy();
    }

    long process(HarvestJobImpl job, int maxNetworkTries)
            throws HarvesterException {
        if (job.isState(State.CANCELED)) {
            return -1;
        }
        if (!job.isState(State.RUNNING)) {
            throw new HarvesterException("job is not in running state");
        }

        boolean ok;
        try {
            switch (job.getTask()) {
            case IDENTIFY_REPOSITORY:
                ok = doIdentifyTask(job);
                if (ok) {
                    job.setTask(Task.ENUMERATE_METADATA_FORMATS);
                }
                break;
            case ENUMERATE_METADATA_FORMATS:
                ok = doListMetadataFormatsTask(job);
                if (ok) {
                    job.setTask(Task.HARVEST_RECORDS);
                }
                break;
            case HARVEST_RECORDS:
                // check for new prefix and notify start, if needed
                if (job.isHarvestingNewPrefix()) {
                    job.onStartListingRecords();
                }
                ok = doListRecordsTask(job);
                if (ok) {
                    // notify end of prefix
                    job.onFinishListingRecords();
                    if (!job.removeCurrentPrefix()) {
                        job.setState(State.DONE);
                    }
                } else if ((job.getResumptionToken() != null) && (delay == 0)) {
                    /*
                     * special case: request successful, but list was split
                     * die to OAI flow control
                     */
                    ok = true;
                }
                break;
            default:
                throw new HarvesterException("internal error");
            } // switch

            if (ok) {
                job.resetNetworkRequestCount();
            } else {
                if (delay > 0) {
                    job.setNetworkRequestDelay(delay);
                }
                if (job.incNetworkRequestCount() >= maxNetworkTries) {
                    throw new HarvesterException("too many network retries: " +
                            job.getNetworkRequestCount());
                }
            }
            return delay;
        } catch (Throwable e) {
            job.setState(State.ERROR);
            if (e instanceof HarvesterException) {
                throw (HarvesterException) e;
            } else if (e instanceof XMLStreamException) {
                throw new HarvesterException("malformed OAI response", e);
            } else if (e instanceof IOException) {
                throw new HarvesterException("network error", e);
            } else {
                throw new HarvesterException("unexpected exception", e);
            }
        }
    }

    private boolean doIdentifyTask(HarvestJobImpl job) throws IOException,
            XMLStreamException, HarvesterException {
        Response response = null;
        try {
            uriBuilder.setVerb(VERB_IDENTIFY);
            response = execute(job, false);
            if (response == null) {
                return false;
            }

            response.readStart(OAI_NS, "Identify", true);

            String s = response.readContent(OAI_NS, "repositoryName", true);
            job.setRepositoryName(s);

            s = response.readContent(OAI_NS, "baseURL", true);
            job.setBaseURL(s);

            s = response.readContent(OAI_NS, "protocolVersion", true);
            if (!PROTOCOL_VERSION.equals(s)) {
                throw new HarvesterException("unsupported protocol version: " +
                        s);
            }
            job.setProtocolVersion(PROTOCOL_VERSION);

            List<String> adminEmail = null;
            do {
                s = response.readContent(OAI_NS, "adminEmail",
                        (adminEmail == null));
                if (s != null) {
                    if (adminEmail == null) {
                        adminEmail = new ArrayList<String>();
                    }
                    adminEmail.add(s);
                }
            } while (s != null);
            job.setAdminEmail(adminEmail);

            s = response.readContent(OAI_NS, "earliestDatestamp", true);
            job.setEarliestTimestamp(s);

            s = response.readContent(OAI_NS, "deletedRecord", true);
            if (DELETED_NOTION_NO.equals(s)) {
                job.setDeletedNotion(DeletedNotion.NO);
            } else if (DELETED_NOTION_PERSISTENT.equals(s)) {
                job.setDeletedNotion(DeletedNotion.PERSISTENT);
            } else if (DELETED_NOTION_TRANSIENT.equals(s)) {
                job.setDeletedNotion(DeletedNotion.TRANSIENT);
            } else {
                throw new HarvesterException(
                        "invalid value for 'deletedRecord': " + s);
            }

            s = response.readContent(OAI_NS, "granularity", true);
            if (GRANULARITY_DAYS.equals(s)) {
                job.setGranularity(Granularity.DAYS);
            } else if (GRANULARITY_SECONDS.equals(s)) {
                job.setGranularity(Granularity.SECONDS);
            } else {
                throw new HarvesterException(
                        "invalid value for 'granularity': " + s);
            }

            int mask = 0;
            do {
                s = response.readContent(OAI_NS, "compression", false);
                if (s != null) {
                    if (COMPRESSION_DEFLATE.equals(s)) {
                        mask |= Repository.COMPRESSION_METHOD_DEFLATE;
                    } else if (COMPRESSION_GZIP.equals(s)) {
                        mask |= Repository.COMPRESSION_METHOD_GZIP;
                    }
                }
            } while (s != null);
            job.setCompressionMask(mask);

            while (response.readStart(OAI_NS, "description", false)) {
                response.consumeWhitespace();
                final String namespaceURI = response.readNamespaceURI();
                System.err.println("DESCRIPTION: URI = " + namespaceURI);
                response.readEnd(OAI_NS, "description", true);
            }
            response.readEnd(OAI_NS, "Identify");
            response.close();
            
            job.onIdentify();

            return true;
        } finally {
            if (response != null) {
                response.release(job);
            }
        }
    }

    private boolean doListMetadataFormatsTask(HarvestJobImpl job)
            throws IOException, XMLStreamException, HarvesterException {
        Response response = null;
        try {
            uriBuilder.setVerb(VERB_LIST_METADATAFORMATS);
            response = execute(job, true);
            if (response == null) {
                return false;
            }

            response.readStart(OAI_NS, "ListMetadataFormats", true);
            List<MetadataFormat> metadataFormats =
                new ArrayList<MetadataFormat>();
            while (response.readStart(OAI_NS, "metadataFormat",
                    metadataFormats.isEmpty())) {
                final String prefix =
                    response.readContent(OAI_NS, "metadataPrefix", true);
                final String schema =
                    response.readContent(OAI_NS, "schema", true);
                final String namespace =
                    response.readContent(OAI_NS, "metadataNamespace", true);
                metadataFormats.add(
                        new MetadataFormatImpl(prefix, schema, namespace));
                response.readEnd(OAI_NS, "metadataFormat");
            } // while
            response.readEnd(OAI_NS, "ListMetadataFormats");
            response.close();

            job.onListMetadataFormats(metadataFormats);
            
            return true;
        } finally {
            if (response != null) {
                response.release(job);
            }
        }
    }

    private boolean doListRecordsTask(HarvestJobImpl job) throws IOException,
            XMLStreamException, HarvesterException {
        Response response = null;
        try {
            String token = job.getResumptionToken();
            uriBuilder.setVerb(VERB_LIST_RECORDS);
            if (token == null) {
                uriBuilder.addParameter(ARG_METADATA_PREFIX,
                        job.getCurrentPrefix());
                if (job.getFrom() != null) {
                    uriBuilder.addParameter("from",
                            formatDate(job.getGranularity(), job.getFrom()));
                }
                if (job.getUntil() != null) {
                    uriBuilder.addParameter("until",
                            formatDate(job.getGranularity(), job.getUntil()));
                }
                if (job.getSet() != null) {
                    uriBuilder.addParameter("set", job.getSet());
                }
            } else {
                uriBuilder.addParameter(ARG_RESUMPTION_TOKEN, token);
                job.incResumptionCount();
            }

            response = execute(job, true);
            if (response == null) {
                return false;
            }

            String nextToken = null;

            response.readStart(OAI_NS, "ListRecords", true);
            header.reset();
            boolean first = true;
            while (response.readStart(OAI_NS, "record", first)) {
                response.readStart(OAI_NS, "header", true, true);
                String s = response.readAttributeValue(null, "status");
                if (s != null) {
                    if ("deleted".equals(s)) {
                        header.setDeleted(true);
                    } else {
                        throw new XMLStreamException("attribute 'status' of " +
                                "element 'record' must contain 'deleted'");
                    }
                }
                response.consumeStart();
                s = response.readContent(OAI_NS, "identifier", true);
                header.setIdentifier(s);
                s = response.readContent(OAI_NS, "datestamp", true);
                header.setDatestamp(parseDate(job.getGranularity(), s));
                while (response.readStart(OAI_NS, "setSpec", false)) {
                    header.addSetSpec(response.readString(true));
                    response.readEnd(OAI_NS, "setSpec", true);
                }
                response.readEnd(OAI_NS, "header");
                job.onRecord(header);
                if (response.readStart(OAI_NS, "metadata", true)) {
                    response.consumeWhitespace();
                    readerWrapper.reset(response.getXMLStreamReader());
                    job.onRecordMetadata(header, readerWrapper);
                    response.consumeWhitespace();
                    response.readEnd(OAI_NS, "metadata");
                }
                while (response.readStart(OAI_NS, "about", false)) {
                    response.consumeWhitespace();
                    job.onRecordAbout(header, null);
                    response.consumeWhitespace();
                    response.readEnd(OAI_NS, "about", true);
                } // while (about)
                response.readEnd(OAI_NS, "record");
                first = false;
                header.reset();
            } // while (record)

            /*
             * expect resumptionToken (possibly empty) when request was resumed
             */
            if (response.readStart(OAI_NS, "resumptionToken",
                    (token != null), true)) {
                // XXX: read attributes here!
                response.consumeStart();

                nextToken = response.readString(false);
                if (nextToken != null) {
                    job.setResumptionToken(nextToken);
                }
                response.readEnd(OAI_NS, "resumptionToken");
            }
            response.readEnd(OAI_NS, "ListRecords");
            
            response.close();

            return (nextToken == null);
        } finally {
            if (response != null) {
                response.release(job);
            }
        }
    }

    private Response execute(HarvestJobImpl job, boolean compress)
            throws IOException, XMLStreamException, HarvesterException {
        long now = System.currentTimeMillis();
        job.incRequestCount();
        final URI uri = uriBuilder.createURI(job.getRepositoryURI());
        System.err.println("HTTP GET " + uri);
        final HttpGet request = new HttpGet(uri);
        if (compress) {
            if ((job.getCompressionMask() &
                    Repository.COMPRESSION_METHOD_GZIP) > 0) {
                request.addHeader(HEADER_ACCEPT_ENCODING, "gzip");
            }
            if ((job.getCompressionMask() &
                    Repository.COMPRESSION_METHOD_DEFLATE) > 0) {
                request.addHeader(HEADER_ACCEPT_ENCODING, "deflate");
            }
            request.addHeader(HEADER_ACCEPT_ENCODING, "identity");
        }
        request.addHeader(HEADER_CONNECTION, "close");
        // FIXME: get timeout values from somewhere?
        HttpParams params = request.getParams();
        params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 180000);
        params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 30000);

        final HttpResponse response = client.execute(request);
        job.addToNetworkTime(System.currentTimeMillis() - now);
        int status = response.getStatusLine().getStatusCode();
        if (status == HttpStatus.SC_OK) {
            delay = 0;
            return new Response(job, response.getEntity(), harvester);
        } else {
            request.abort();
            if (status == HttpStatus.SC_SERVICE_UNAVAILABLE) {
                System.err.println("-> " + status);
                delay = parseTryAfter(response);
            } else {
                throw new HarvesterException(
                        "provider returned unexpected HTTP status: " + status);
            }
        }
        return null;
    }

    private long parseTryAfter(HttpResponse response) {
        final Header s = response.getFirstHeader("Retry-After");
        if (s != null) {
            final String v = s.getValue();
            try {
                return Long.parseLong(v);
            } catch (NumberFormatException e) {
                /* IGNORE */
            }
        }
        return 1000;
    }

    private Date parseDate(Repository.Granularity granularity, String s)
            throws HarvesterException {
        String[] patterns = null;
        switch (granularity) {
        case DAYS:
            patterns = DATEFORMATS_DAYS;
            break;
        case SECONDS:
            patterns = DATEFORMATS_FULL;
        }
        MutableDateTime date = new MutableDateTime(0);
        for (int i = 0; i < patterns.length; i++) {
            DateTimeFormatter fmt = DateTimeFormat
                .forPattern(patterns[i])
                .withZone(DateTimeZone.UTC);
            if (fmt.parseInto(date, s, 0) == s.length()) {
                if (patterns[i].equals(DATEFORMATS_DAYS[0])) {
                    date.setTime(0, 0, 0, 0);
                }
                return date.toDate();
            }
        }
        throw new HarvesterException("invalid date: " + s);
    }

    private String formatDate(Repository.Granularity granularity, Date date) {
        String pattern = null;
        switch (granularity) {
        case DAYS:
            pattern = DATEFORMATS_DAYS[0];
            break;
        case SECONDS:
            pattern = DATEFORMATS_FULL[0];
        }
        DateTimeFormatter fmt =
            DateTimeFormat.forPattern(pattern).withZone(DateTimeZone.UTC);
        return fmt.print(new DateTime(date));
    }

} // class HarvestWorker
